"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit legends. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import FilteredElementCollector, View, ViewType

from duHast.Revit.Views.views import get_views_not_on_sheet


def get_view_legends(doc):
    """
    Get all view legends in a model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: All view legends in the model
    :rtype: list of Autodesk.Revit.DB.View
    """

    view_legends = []
    col = FilteredElementCollector(doc).OfClass(View)
    for v in col:
        # filter out templates
        if v.ViewType == ViewType.Legend:
            view_legends.append(v)
    return view_legends


def get_view_legends_not_placed(doc):
    """
    Returns all unplaced legend views

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: All view legends in the model not on a sheet
    :rtype: list of Autodesk.Revit.DB.View
    """

    # get all legends in model:
    all_legend_views = get_view_legends(doc=doc)

    # get all views not on sheets
    all_views_not_on_sheets = get_views_not_on_sheet(doc=doc)

    unplaced_legends = []
    # loop over views and find legends
    for view in all_views_not_on_sheets:
        if view.ViewType == ViewType.Legend:
            unplaced_legends.append(view)

    return unplaced_legends



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
.net dll names currently in use.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This is required since the dll's used have a version number in their file name.

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
# 

DLL_VERSION = ".23.0.0.5"

# family reloader ui
FAMILY_RELOADER_UI = "FamilyReloaderUI{}.dll".format(DLL_VERSION)

# pdf and dwg exporter selection ui
PDF_AND_DWG_EXPORTER_SELECTION_UI = "PDFDWGExporterSelectionUI{}.dll".format(DLL_VERSION)

# pdf and dwg exporter settings ui
PDF_AND_DWG_EXPORTER_SETTINGS_UI = "PDFDWGExporterUI{}.dll".format(DLL_VERSION)

# .net revit utility classes
REVIT_UTILS= 'RevitUtils{}.dll'.format(DLL_VERSION)

# .net custom controls
WPF_CUSTOM_CONTROLS = "duHastUICustomControls{}.dll".format(DLL_VERSION)

# .net wrapper class  of csv helper
FILE_WRAPPER = "FileIOWrapper{}.dll".format(DLL_VERSION)

# .net utility classes
UTILITY = "Utils{}.dll".format(DLL_VERSION)


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains the view property data filter.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


def filter_data_by_properties(data, headers, properties, default_headers):
    """
    Filters view data by supplied property names.
    Data gets filtered twice: property needs to exist in headers list as well as in properties list.

    :param data: List of sheet properties to be kept.
    :type data: list of list of str
    :param headers: Filter: list of headers representing property names.
    :type headers: list of str
    :param properties: list of optional properties to be extracted from data
    :type properties: list of str
    :param default_headers: list of view properties always extracted from data
    :type default_headers: list of str
    :return: List of sheet properties matching filters.
    :rtype: list of list of str
    """

    # filter data out
    new_data = []
    for d in data:
        data_row = {}
        for default_head in default_headers:
            if default_head in d:
                data_row[default_head] = d[default_head]
            else:
                data_row[default_head] = "default property does not exist on element"
        for prop in properties:
            if prop in d:
                data_row[prop] = d[prop]
            else:
                data_row[prop] = "property does not exist on element"
        new_data.append(data_row)
    return new_data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A container class used to group ElementParameterFilters.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


Stores view filters rules


"""


#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import json

from duHast.Utilities.Objects import base
from duHast.Revit.Views.Objects.Data.view_filter_rule import ViewFilterRule

class ViewFilterLogicContainer(base.Base):
    def __init__(self, data_type="view filter logic container", j=None, **kwargs):
        """
        Class constructor.
        """

        super(ViewFilterLogicContainer, self).__init__(**kwargs)

        self.data_type = data_type
        self.view_filter_rules = []
        self.logic_containers = []
        self.logic_container_type = ""

        # print("here in ViewFilterLogicContainer init...")
        # print("j: {}".format(j))
        # check if any data was past in with constructor!
        if j is not None:
            # check type of data that came in:
            if type(j) == str:
                # a string
                j = json.loads(j)
            elif type(j) == dict:
                # no action required
                pass
            else:
                raise ValueError(
                    "Argument supplied must be of type string or type dictionary"
                )

            # load values and throw exception if something is missing!
            try:
                # get the container type
                self.logic_container_type = j["logic_container_type"]

                # get rules
                rules = j["view_filter_rules"]
                for r in rules:
                    self.view_filter_rules.append(ViewFilterRule(j=r))

                # get containers
                containers = j["logic_containers"]
                for c in containers:
                    self.logic_containers.append(ViewFilterLogicContainer(j=c))

            except Exception as e:
                raise ValueError(
                    "Node {} failed to initialise with: {}".format(
                        data_type, e
                    )
                )
    

    def __eq__(self, other):
        """
        Equality method to compare two objects of this class.

        :param other: Other object to compare with.
        :type other: ViewFilterLogicContainer

        :return: True if both objects are equal, False otherwise.
        :rtype: bool
        """

        if not isinstance(other, ViewFilterLogicContainer):
            # don't attempt to compare against unrelated types
            return False

        # compare the logic container type
        if self.logic_container_type != other.logic_container_type:
            return False

        if self.data_type != other.data_type:
            return False

        if len(self.view_filter_rules) != len(other.view_filter_rules):
            return False
        else:
            for rule in self.view_filter_rules:
                if rule not in other.view_filter_rules:
                    return False

        if len(self.logic_containers) != len(other.logic_containers):
            return False
        else:
            for container in self.logic_containers:
                if container not in other.logic_containers:
                    return False

        # all tests passed, must be equal
        return True
    

    def __ne__(self, other):
        """
        Inequality method to compare two objects of this class.

        :param other: Other object to compare with.
        :type other: ViewFilterLogicContainer

        :return: True if both objects are not equal, False otherwise.
        :rtype: bool
        """

        return not self.__eq__(other)
    
    def __hash__(self):
        """
        Override the default hash behavior (that returns the id or the object)
        """

        return hash((self.logic_container_type, tuple(self.view_filter_rules), tuple(self.logic_containers)))


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to dimensions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import (
    BuiltInParameter,
    ElementId,
    FilteredElementCollector,
    DimensionType,
    Dimension,
)

from duHast.Revit.Common.parameter_get_utils import get_built_in_parameter_value
from duHast.Revit.Common.common import get_ids_from_element_collector
from duHast.Revit.Annotation.arrow_heads import (
    get_arrow_head_ids_from_type,
    ARROWHEAD_PARAS_DIM,
)


def get_dim_types(doc):
    """
    Gets all dimension types in a model
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of Dimension Types
    :rtype: Autodesk.Revit.DB.FilteredElementCollector of DimensionType
    """

    return FilteredElementCollector(doc).OfClass(DimensionType)


def get_dim_type_ids(doc):
    """
    Gets all dimension type ids in a model
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing Dimension Types
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col = FilteredElementCollector(doc).OfClass(DimensionType)
    ids = get_ids_from_element_collector(col)
    return ids


def get_all_dimension_elements(doc):
    """
    Gets all dimension elements placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of Dimensions
    :rtype: Autodesk.Revit.DB.FilteredElementCollector of Dimension
    """

    return FilteredElementCollector(doc).OfClass(Dimension)


def get_symbol_ids_from_dim_types(doc):
    """
    Gets all family symbol ids used as centre line symbol from all dim styles in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing family symbols
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    dim_ts = get_dim_types(doc)
    for t in dim_ts:
        id = get_built_in_parameter_value(
            t, BuiltInParameter.DIM_STYLE_CENTERLINE_SYMBOL
        )
        if id not in ids and id != ElementId.InvalidElementId and id != None:
            ids.append(id)
    return ids


def get_dim_type_arrow_head_ids(doc):
    """
    Gets all arrow head symbol ids used in dim types in a model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing arrow head symbols
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    used_ids = get_arrow_head_ids_from_type(doc, get_dim_types, ARROWHEAD_PARAS_DIM)
    return used_ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
data to revit api FilterValueRule conversion helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#
from System import Enum

from duHast.Revit.Views.Objects.Data.view_filter_rule import ViewFilterRule
from duHast.Revit.Common.parameter_project import get_project_parameter_definition_by_name
from duHast.Revit.SharedParameters.shared_parameters import get_shared_parameter_by_guid
from duHast.Revit.Views.Utility.convert_data_to_filter_value_provider import get_value_provider_class
from duHast.Revit.Views.Utility.convert_data_to_filter_evaluator import get_evaluator_class


from duHast.Utilities.Objects.result import Result


from Autodesk.Revit.DB import  (
    BuiltInParameter,
    ElementId, 
    FilterDoubleRule,
    FilterIntegerRule,
    FilterInverseRule,
    FilterElementIdRule,
    FilterStringRule,
    FilterGlobalParameterAssociationRule
)

# dictionary containing varies rule mappings

# numeric rules (all of these inherit from: Autodesk.Revit.DB.FilterNumericValueRule)
# string rules (just one : Autodesk.Revit.DB.FilterStringRule)

class_mapping = {
    "FilterDoubleRule": FilterDoubleRule,
    "FilterElementIdRule":  FilterElementIdRule,
    "FilterIntegerRule": FilterIntegerRule,
    "FilterStringRule":  FilterStringRule,
    "FilterGlobalParameterAssociationRule": FilterGlobalParameterAssociationRule,

}

# notes: rules require 
# - a rule type (see class_mapping above)
# - a parameter
# - a value provider (ParameterValueProvider Class) 
# - a rule value


def create_filter_double_rule(evaluator, value_provider, rule_data_instance):
    """
    Creates a FilterDoubleRule.
    
    :param evaluator: The evaluator for the rule.
    :type evaluator: Autodesk.Revit.DB.FilterDoubleEvaluator
    :param value_provider: The value provider for the rule.
    :type value_provider: Autodesk.Revit.DB.ParameterValueProvider
    :param rule_data_instance: The rule data instance containing the rule value and epsilon.

    :return: The created FilterDoubleRule or None if creation failed.
    """

    try:
        rule = FilterDoubleRule( value_provider, evaluator, float(rule_data_instance.rule_value), float(rule_data_instance.epsilon))
        return rule
    except Exception:
        pass
    return None


def create_filter_integer_rule(evaluator, value_provider, rule_data_instance):
    """
    Creates a FilterIntegerRule.

    :param evaluator: The evaluator for the rule.
    :type evaluator: Autodesk.Revit.DB.FilterIntegerEvaluator
    :param value_provider: The value provider for the rule.
    :type value_provider: Autodesk.Revit.DB.ParameterValueProvider
    :param rule_data_instance: The rule data instance containing the rule value.

    :return: The created FilterIntegerRule or None if creation failed.
    """

    try:
        rule = FilterIntegerRule( value_provider, evaluator, int(rule_data_instance.rule_value))
        return rule
    except Exception:
        pass
    return None


def create_filter_element_id_rule(evaluator, value_provider, rule_data_instance):
    """
    Creates a FilterElementIdRule.

    :param evaluator: The evaluator for the rule.
    :type evaluator: Autodesk.Revit.DB.FilterIntegerEvaluator
    :param value_provider: The value provider for the rule.
    :type value_provider: Autodesk.Revit.DB.ParameterValueProvider
    :param rule_data_instance: The rule data instance containing the rule value.

    :return: The created FilterElementIdRule or None if creation failed.
    """

    try:
        rule = FilterElementIdRule( value_provider, evaluator, ElementId(int(rule_data_instance.rule_value)))
        return rule
    except Exception:
        pass
    return None


def create_filter_string_rule(evaluator, value_provider, rule_data_instance):
    """
    Creates a FilterStringRule.

    :param evaluator: The evaluator for the rule.
    :type evaluator: Autodesk.Revit.DB.FilterStringEvaluator
    :param value_provider: The value provider for the rule.
    :type value_provider: Autodesk.Revit.DB.ParameterValueProvider
    :param rule_data_instance: The rule data instance containing the rule value.

    :return: The created FilterStringRule or None if creation failed.
    """

    rule = None
    try:
        rule = FilterStringRule( value_provider, evaluator, rule_data_instance.rule_value)
    except Exception:
        pass
    return rule


def create_filter_global_parameter_association_rule(evaluator, value_provider, rule_data_instance):
    """
    Creates a FilterGlobalParameterAssociationRule.

    :param evaluator: The evaluator for the rule.
    :type evaluator: Autodesk.Revit.DB.FilterGlobalParameterAssociationEvaluator
    :param value_provider: The value provider for the rule.
    :type value_provider: Autodesk.Revit.DB.ParameterValueProvider
    :param rule_data_instance: The rule data instance containing the rule value.

    :return: The created FilterGlobalParameterAssociationRule or None if creation failed.
    """

    rule = None
    try:
        rule = FilterGlobalParameterAssociationRule( value_provider, evaluator, ElementId(int(rule_data_instance.rule_value)))
    except Exception:
        pass
    return rule


def get_rule_parameter(doc, rule_data_instance):
    """
    Gets the parameter for the given rule data instance.

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param rule_data_instance: The rule data instance containing the parameter information.
    :type rule_data_instance: ViewFilterRule
    :return: A Result object containing the parameter or an error message.
    """

    return_value = Result()
   
    # get the parameter
    # if the id is negative it is a built in parameter id, easiest since no further checking is required
    # if the id is positive but no guid is provided it is a project parameter, try to match by name only
    # if the id is positive and a guid is provided it is a shared parameter, try to match by guid only

    if rule_data_instance.parameter_id < 0:
        # built in parameter
        try:
            built_in_param = Enum.Parse(BuiltInParameter, rule_data_instance.parameter_name)
            return_value.result.append(built_in_param)
            return_value.append_message( "Successfully got built in parameter: {}".format(rule_data_instance.parameter_name))

        except Exception as e:
            return_value.update_sep(False, "Failed to get built in parameter: {} with exception: {}".format( rule_data_instance.parameter_name, e))
        return return_value
    
    elif rule_data_instance.parameter_id > 0 and rule_data_instance.parameter_guid == "":
        # project parameter, try to match by name only
        try:
            parameter_definition = get_project_parameter_definition_by_name(doc, rule_data_instance.parameter_name)
            if parameter_definition:
                return_value.result.append(parameter_definition)
                return_value.append_message( "Successfully got project parameter: {}".format(rule_data_instance.parameter_name))
            else:
                return_value.update_sep(False, "Failed to get project parameter: {}. No matching parameter found by name.".format( rule_data_instance.parameter_name))
        except Exception as e:
            return_value.update_sep(False, "Failed to get project parameter: {} with exception: {}".format( rule_data_instance.parameter_name, e))
        return return_value
    elif rule_data_instance.parameter_id > 0 and rule_data_instance.parameter_guid != "":
        # shared parameter, try to match by guid only
        try:
            parameter_definition = get_shared_parameter_by_guid(doc, rule_data_instance.parameter_guid)
            if parameter_definition:
                return_value.result.append(parameter_definition)
                return_value.append_message( "Successfully got shared parameter: {}[{}]".format(rule_data_instance.parameter_name, rule_data_instance.parameter_guid))
            else:
                return_value.update_sep(False, "Failed to get shared parameter: {}. No matching parameter found by guid: {}".format( rule_data_instance.parameter_name, rule_data_instance.parameter_guid))
        except Exception as e:
            return_value.update_sep(False, "Failed to get shared parameter: {} with exception: {}".format( rule_data_instance.parameter_name, e))
        return return_value
    else:
        # no valid id provided, shouldnt get here but just in case
        return_value.update_sep(False, "Failed to get parameter: {}. No valid parameter id or guid provided.".format( rule_data_instance.parameter_name))
        return return_value


def get_rule_parameter_id(doc, rule_data_instance, parameter ):
    
    # decide on which parameter id to use
    parameter_id = ElementId.InvalidElementId

    if rule_data_instance.parameter_id < 0:
        # built in parameter, use what has been stored in the rule
        parameter_id = ElementId(rule_data_instance.parameter_id)
    elif rule_data_instance.parameter_id > 0:
        # use the past in parameter
        parameter_id = parameter.Id
    
    return parameter_id


def convert_data_to_rule(doc, rule_data_instance):
    """
    Converts a ViewFilterRule data instance to a Revit FilterValueRule.

    Args:
        doc (Revit Document): The Revit document object.
        rule_data_instance (ViewFilterRule): An instance of the ViewFilterRule class containing the rule data.

    Returns:
        FilterValueRule or None: The created FilterValueRule object if successful, otherwise None.
    """
    return_value = return_value = Result()

    # check input
    if not isinstance(rule_data_instance, ViewFilterRule):
        return return_value
    
    # get the parameter
    parameter_result = get_rule_parameter(doc, rule_data_instance)

    # get out if something went wrong
    if not parameter_result.status:
        return_value.update_sep(False, "Failed to get parameter for rule: {}. Error: {}".format(rule_data_instance.parameter_name, parameter_result.message))
        return return_value

    # get the parameter ( or parameter definition)
    parameter = parameter_result.result[0]
    
    # decide on which parameter id to use
    parameter_id = get_rule_parameter_id(doc, rule_data_instance, parameter)
    #print(parameter_id)

    # get the value provider class
    value_provider_class = get_value_provider_class(rule_data_instance.value_provider)

    if not value_provider_class:
        return_value.update_sep(False, "Failed to get value provider class: {} for rule: {}".format(rule_data_instance.value_provider, rule_data_instance.parameter_name))
        return return_value

    # create the value provider
    value_provider = value_provider_class(parameter_id)
    #print(value_provider)

    # get the evaluator class
    evaluator_class = get_evaluator_class(rule_data_instance.evaluation_type)

    if not evaluator_class:
        return_value.update_sep(False, "Failed to get evaluator class: {} for rule: {}".format(rule_data_instance.evaluation_type, rule_data_instance.parameter_name))
        return return_value
    
    # convert the rule value to the correct type
    evaluator = evaluator_class()
    #print(evaluator)

    # init rule
    rule = None
    # create the rule
    if rule_data_instance.rule_type == FilterDoubleRule.__name__:
        # double rules need a double as rule value and an epsilon value

        rule = create_filter_double_rule( evaluator, value_provider, rule_data_instance)
        return_value.append_message("Created double rule: {} with epsilon: {}".format(rule_data_instance.rule_value, rule_data_instance.epsilon))
    elif rule_data_instance.rule_type == FilterIntegerRule.__name__:
        # integer rules need an integer as rule value

        rule = create_filter_integer_rule(evaluator, value_provider, rule_data_instance)
        return_value.append_message("Created integer rule: {}".format(rule_data_instance.rule_value))
    elif rule_data_instance.rule_type == FilterElementIdRule.__name__:
        # element id rules need an integer as rule value
        
        rule = create_filter_element_id_rule(evaluator, value_provider, rule_data_instance)
        return_value.append_message("Created element id rule: {}".format(rule_data_instance.rule_value))
    elif rule_data_instance.rule_type == FilterStringRule.__name__:
        # string rule
        
        rule = create_filter_string_rule(evaluator, value_provider, rule_data_instance)
        return_value.append_message("Created string rule: {}".format(rule_data_instance.rule_value))
    elif rule_data_instance.rule_type == FilterGlobalParameterAssociationRule.__name__:
        # global parameter association rule
        
        rule = create_filter_global_parameter_association_rule(evaluator, value_provider, rule_data_instance)
        return_value.append_message("Created global parameter association rule: {}".format(rule_data_instance.rule_value))
    else:
        # uh this is bad
        return_value.update_sep(False, "Failed to create rule. Unknown rule type: {} for rule: {}".format(rule_data_instance.rule_type, rule_data_instance.parameter_name))
        return return_value

    # check if rule was created
    if not rule:
        return_value.update_sep(False, "Failed to create rule of type: {} for rule: {}".format(rule_data_instance.rule_type, rule_data_instance.parameter_name))
        return return_value
    
    # check if inverted rule?
    if rule_data_instance.is_inversed:
        rule = FilterInverseRule(rule)
        return_value.append_message("Inverted rule as requested.")
    
    # return the rule
    return_value.result.append(rule)

    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to directory IO . 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import os
from os import path
import os.path
import shutil
import uuid

# net import for directory delete with fall back
from System.IO import Directory, IOException
from System.Threading import Thread

def is_directory(directory_path):
    """
    Check if a given path is a directory
    :param directory_path: Fully qualified directory path
    :type directory_path: str
    :return: True if directory, otherwise False
    :rtype: bool
    """
    try:
        return os.path.isdir(directory_path)
    except Exception:
        return False


def directory_empty_delete(full_directory_path):
    """
    Deletes an empty directory
    :param full_directory_path: Path to directory
    :type full_directory_path: str
    :return: True directory deleted, otherwise False
    :rtype: bool
    """

    try:
        os.rmdir(full_directory_path)
        value = True
    except Exception:
        value = False
    return value


def directory_delete(full_directory_path):
    """
    Deletes a directory (even if it contains files)

    :param full_directory_path: Path to directory
    :type full_directory_path: str
    :return: True directory deleted, otherwise False
    :rtype: bool
    """

    try:
        shutil.rmtree(full_directory_path)
        return  True
    except Exception as e:
        return  False


def directory_delete_with_fallback(full_directory_path, max_attempts=3, delay_ms=100):
    """
    Deletes a directory (even if it contains files)
    If it fails, it will try again after a short wait

    :param full_directory_path: Path to directory
    :type full_directory_path: str
    :param max_attempts: Number of attempts to delete the directory
    :type max_attempts: int
    :param delay_ms: Delay in milliseconds between attempts
    :type delay_ms: int
    :return: True directory deleted, otherwise False
    :rtype: bool
    """

    for attempt in range(max_attempts):
        try:
            if Directory.Exists(full_directory_path):
                # delete if it exists
                Directory.Delete(full_directory_path, True)
                return True
            else:
                # already gone
                return True
        except IOException as e:
            if attempt < max_attempts - 1:
                # wait a bit before retrying
                Thread.Sleep(delay_ms)
            else:
                # last attempt failed
                return False
        except Exception as e:
            # something went wrong
            return False




def get_child_directories(full_directory_path):
    """
    Returns the immediate subdirectories of directory
    
    :param full_directory_path: Path to directory
    :type full_directory_path: str
    :return: any sub directories, empty list if none exist
    :rtype: list of str
    """

    sub_folders_with_paths = []
    for root, dirs, files in os.walk(full_directory_path):
        for dir in dirs:
            sub_folders_with_paths.append(os.path.join(root, dir))
        break
    return sub_folders_with_paths


def get_all_nested_directories(full_directory_path):
    # returns all nested direcory paths
    """
    Returns all nested directories of directory

    :param full_directory_path: Path to directory
    :type full_directory_path: str
    :return: any sub directories, empty list if none exist
    :rtype: list of str
    """

    nested_directories = []
    for entry in os.listdir(full_directory_path):
        full_path = os.path.join(full_directory_path, entry)
        if os.path.isdir(full_path):
            nested_directories.append(full_path)
            nested_directories.extend(get_all_nested_directories(full_path))
    return nested_directories


def get_parent_directory(full_directory_path):
    """
    Returns the parent directory of directory, or empty string if invalid directory
    :param full_directory_path: Path to directory
    :type full_directory_path: str
    :return: parent directory, or empty string
    :rtype: str
    """

    parent_dir = ""
    try:
        parent_dir = os.path.dirname(full_directory_path)
    except Exception:
        pass
    return parent_dir


def create_directory(root, folder_name):
    """
    Create a folder.
    :param root: Directory path in which the new folder is to be created
    :type root: str
    :param folder_name: New folder name.
    :type folder_name: str
    :return: True if folder is created, otherwise False
    :rtype: bool
    """

    dir_name = path.join(root, folder_name)
    flag = True
    try:
        os.mkdir(dir_name)
    except Exception as e:
        # just in case the folder does exist (created by another instance at almost the same time)
        if "[Errno 17]" not in str(e):
            flag = False
    return flag


def create_nested_directory(directory_path):
    """
    Creates a directory and all necessary subdirectories.

    :param directory_path: Fully qualified directory path
    :type directory_path: str
    
    :return: True if directory is created, otherwise False
    :rtype: bool
    """
    try:
        import clr
        clr.AddReference("System")
        from System.IO import Directory

        # Create the directory and all necessary subdirectories
        Directory.CreateDirectory(directory_path)

        return True
    except Exception:

        return False


def create_target_directory(root_path, folder_name):
    """
    Create a folder.
    Checks whether folder exists and if not attempts to create it.

    :param root: Directory path in which the new folder is to be created
    :type root: str
    :param folder_name: New folder name.
    :type folder_name: str
    :return: True if folder is created, otherwise False
    :rtype: bool
    """

    # check if folder exists
    flag = True
    if path.exists(root_path + "\\" + folder_name) == False:
        # create new folder
        flag = create_directory(root_path, folder_name)
    return flag


def directory_exists(directory_path):
    """
    Check if a given directory exists
    :param directory_path: Fully qualified directory path
    :type directory_path: str
    :return: True if directory exists, otherwise False
    :rtype: bool
    """
    if path.exists(directory_path):
        return True
    else:
        return False



def get_current_user_documents_directory():
    """
    Get the current user's documents directory.

    :return: Path to the user's documents directory.
    :rtype: str
    """
    
    import clr
    clr.AddReference("System")
    from System import Environment

    # Get the user's Documents folder path
    documents_path = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments)

    return documents_path


def create_temp_directory(root_directory=None):
    """
    Creates a temporary directory in the system temp folder.

    :param root_directory: Root directory to create the temp folder in. If None, uses the system temp folder.
    :type root_directory: str or None. If none provided, the user's documents folder is used.
    :return: Path to the created temporary directory.
    :rtype: str
    """
    
    # Generate a random filename
    random_filename = uuid.uuid4().hex  # Creates a unique random string

    # If no root directory is provided, use the users doc folder
    if root_directory is None:
        root_directory = get_current_user_documents_directory()
    
    temp_directory = os.path.join(root_directory, random_filename)
    # Check if the directory exists before creating
    if not os.path.exists(temp_directory):
        os.makedirs(temp_directory)
        
    
    return temp_directory



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Implementation of a user pick action with a custom element selection filter.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.UI.Selection import ObjectType

from duHast.Revit.UI.Objects.CustomSelectionFilter import CustomSelectionFilter
from duHast.Revit.UI.custom_selection_filter_element import selection_filter_floors
from duHast.Utilities.Objects.result import Result


def get_user_selection(doc, uidoc, ui_text, selection_filter=selection_filter_floors):
    """
    Get user selection of set down families.

    :param doc: Revit Document
    :type doc: Autodesk.Revit.DB.Document
    :param uidoc: Revit UIDocument
    :type uidoc: Autodesk.Revit.UI.UIDocument
    :param ui_text: User text for the selection dialog
    :type ui_text: str
    :return: Tuple containing a list of selected elements and a list of selected element ids.
    :rtype: (list, list)
    """

    return_value = Result()
    elements_selected = []
    elements_selected_ids = []

    # wrap in try catch since pressing cancel in the Revit UI will raise an exception
    try:
        # get the selected elements
        sel_filter = CustomSelectionFilter(selection_filter)
        sel_elements = uidoc.Selection.PickObjects(
            ObjectType.Element, sel_filter, ui_text
        )

        # get the actual selected elements ( PickObject returns a bunch of reference objects only!)
        for e in sel_elements:
            elements_selected_ids.append(e.ElementId)
            element = doc.GetElement(e.ElementId)
            elements_selected.append(element)

        return_value.result = [elements_selected, elements_selected_ids]
    except Exception as e:
        return_value.update_sep(
            False, "User selection failed with exception: {}".format(e)
        )
    return return_value



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#
import os

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_xml import get_all_xml_files_from_directories
from duHast.Revit.Family.family_types_get_data_from_xml import get_family_type_data_from_library
from duHast.Revit.Family.Data.Objects.family_directive_swap_instances_of_type import FamilyDirectiveSwap

def get_family_data_from_file(libraryPath):
    """
    Extracts family type data from XML files in the specified library path.

    :return:
        Result class instance.

        - result.status: XML conversion status will be returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain which xml file was read and converted into family type data.
        - result.result will be [:class:`FamilyTypeDataStorageManager`]

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message
        - Reload.result will be an empty list

    :rtype: :class:`.Result`
    """
    xml_files =  get_all_xml_files_from_directories([libraryPath])

    if len(xml_files) == 0:
        return None
    
    family_data_result = get_family_type_data_from_library(xml_files)
    
    return family_data_result.result


def find_matches_by_key (family_target_data, family_source_data, library_target_key_parameter_name, library_source_key_parameter_name,output):

    return_value = Result()

    key_to_families_mapper = {}

    # loop over source library data and check if key parameter exists in source family 
    for family_source in family_source_data:

        # check if  FamilyTypeDataStorageManager has the key parameter
        for fam_source_type in family_source.family_type_data_storage:

            match_found = False

            key_source_parameter = fam_source_type.get_parameter_by_name( library_source_key_parameter_name)
            if key_source_parameter is None:
                output("Key parameter '{}' not found in source family '{}' type: {}.".format(library_source_key_parameter_name, family_source.family_name, fam_source_type.family_type_name))
                continue

            # get the key value from the source family type
            source_key_value =  key_source_parameter.value
        
            # loop over target library data and check if key parameter exists in target family
            for family_target in family_target_data:
                # check if  FamilyTypeDataStorageManager has the key parameter
                for fam_target_type in family_target.family_type_data_storage:
                    key_target_parameter = fam_target_type.get_parameter_by_name(library_target_key_parameter_name)

                    if key_target_parameter is None:
                        output("Key parameter '{}' not found in target family '{}' type: {}.".format(library_target_key_parameter_name, family_target.family_name, fam_target_type.family_type_name))
                        continue
                
                    # get the key value from the source family type
                    target_key_value =  key_target_parameter.value
                
                    # if the key values match add to result
                    if source_key_value == target_key_value:
                        key_to_families_mapper[source_key_value] = (fam_source_type, fam_target_type)
                        match_found = True
                        break  # break to avoid multiple matches for the same source key value
                
                # get out of this loop too
                if match_found:
                    break

    
    return_value.update_sep(True, "Found {} matches by key.".format(len(key_to_families_mapper)))
    return_value.result.append(key_to_families_mapper)
    return return_value


def map_libraries_by_keys(library_target_report_path, library_source_report_path, library_target_key_parameter_name, library_source_key_parameter_name,output):

    """
    Maps libraries by keys from the target and source report paths.

    :param library_target_report_path: Path to the target library report.
    :param library_source_report_path: Path to the source library report.
    :return: A dictionary mapping keys to their corresponding libraries.
    """
    
    return_value = Result()

    if not os.path.exists(library_target_report_path):
        return_value.update_sep(False, "Target report path does not exist: {library_target_report_path}".format(library_target_report_path=library_target_report_path))
        return return_value

    if not os.path.exists(library_source_report_path):
        return_value.update_sep(False, "Source report path does not exist: {library_source_report_path}".format(library_source_report_path=library_source_report_path))
        return return_value

    # Initialize a dictionary to hold the mapping
    

    family_data_target = get_family_data_from_file(library_target_report_path)
    if (family_data_target is None):
        return_value.update_sep(False, "No family data found in the specified library path.")
        output("No family data found in the specified library path: {}".format(library_target_report_path))
        return return_value

    family_data_source = get_family_data_from_file(library_source_report_path)
    if (family_data_source is None):
        return_value.update_sep(False, "No family data found in the specified library path.")
        output("No family data found in the specified library path: {}".format(library_source_report_path))
        return return_value
    

    library_map_result = find_matches_by_key(family_data_target, family_data_source, library_target_key_parameter_name, library_source_key_parameter_name, output)
    if (library_map_result.status is False):
        return_value.update_sep(False, "Error finding matches by key: {}".format(library_map_result.message))
        return return_value
    
    # get the result dictionary from the library_map_result
    library_map = library_map_result.result[0]

    if len(library_map) == 0:
        return_value.update_sep(False, "No matches found between the libraries.")
        output("No matches found between the libraries.")
        return return_value

    output("Found {} matches between the libraries.".format(len(library_map)))

    return_value.result.append(library_map)
    return_value.update_sep(True, "Libraries mapped successfully.")
    return return_value


def convert_mapping_to_swap_directives(mapping):
    """
    Converts the mapping of libraries to swap directives.

    :param mapping: The mapping of libraries.
    :return: A list of swap directives.
    """
    swap_directives = []
    
    for key, (source_family, target_family) in mapping.items():
        swappy = FamilyDirectiveSwap(
            name=source_family.family_name,
            category=source_family.root_category_path,
            source_type_name=source_family.family_type_name,
            target_family_name=target_family.family_name,
            target_family_type_name=target_family.family_type_name
        )
        swap_directives.append(swappy)
    
    if len(swap_directives) == 0:
        raise Exception("No swap directives created from the mapping.")
    

    
    return swap_directives


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit families helper functions retrieving duplicated families.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A duplicated family is a family which has the name of another family with an appended (1) or (2) etc.
Revit seems to do this when loading families with the same name into a project and the family loaded in has the same name as one already in the project but is slightly different. (May hav to do with versioning of families but not sure yet).


Samples:
NRS_Button_NurseCall_Emergency_ITCL-004
NRS_Button_NurseCall_Emergency_ITCL1

Last characters of a family are removed if they represent a none alphabetical character and are replaced with a 1 or 2 etc to indicate a duplicate family.

LevelHead
LevelHead1

If there are no numbers at the end of the family name then a 1 or 2 etc is appended to indicate a duplicate family.

To identify a duplicated family:
1. check if the last character is a number
2. if so, check if the character before is an alphabetical character, if not ignore ( assume we dont have more than 9 duplicates in a file... )
3. if so, remove the number and check if the remaining string is a match with another family name in the project.
    A match is defined as either
        - the remaining string being the same as another family name of the same Revit category or 
        - the remaining string being the same as another family, of the same Revit category, followed by none alphabetical characters ( make sure its not the same family )




A second part of this module is to find matching types for two given families by their type names. This is useful when replacing a family with another one and wanting to keep the same type (i.e. duplicate families).

When creating duplicate families in a project Revit either keeps the type name the same or appends a " 1" or " 2" etc to the type name.

A match therefore is defined as either:
- the type name being the same as another type name in the other family
- the type name being the same as another type name in the other family followed by 2 none none alphabetical characters (first character is a space, second character is a number)


"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Revit.Family.family_functions import get_category_name_to_family_dict
from Autodesk.Revit.DB import Element

def get_duplicate_family_root_name (family_name):
    """
    Returns the root name of a duplicate family or None if not a duplicate family.

    :param family_name: Name of the family to check.
    :type family_name: str

    :return: Root name of duplicate family or None if not a duplicate family.
    :rtype: str | None
    """

    # return none if this is not a duplicate family ( ie. does not end with a number )
    duplicate_root_name = None

    # check if last character is a number
    if family_name[-1].isnumeric() == False:
        return duplicate_root_name

    # check if the character before is an alphabetical character, if not ignore ( assume we dont have more than 9 duplicates in a file... )
    if family_name[-2].isalpha() == False:
        return duplicate_root_name
    
    # remove the number
    duplicate_root_name = family_name[:-1]
    return duplicate_root_name


def find_original_family_based_on_duplicate_family (original_families_by_category,  original_family_name, duplicate_family_root_name, duplicate_family_category_name):
    """
    Finds the original family based on the duplicate family root name.

    :param original_families_by_category: Dictionary of original families by category.
    :type original_families_by_category: dict[str, list[Autodesk.Revit.DB.Family]]

    :param duplicate_family_root_name: Root name of the duplicate family.
    :type duplicate_family_root_name: str

    :return: The original family or None if not found.
    :rtype: Autodesk.Revit.DB.Family | None
    """

    # check if the remaining string is a match with another family name in the project.
    # A match is defined as either
    # - the remaining string being the same as another family name of the same Revit category or 
    # - the remaining string being the same as another family, of the same Revit category, followed by none alphabetical characters ( make sure its not the same family )

    for family in original_families_by_category[duplicate_family_category_name]:
        if family.Name == duplicate_family_root_name:
            return family
        else:
            # check if the family name starts with the duplicate root name and is followed by none alphabetical characters
            if family.Name.startswith(duplicate_family_root_name) and original_family_name != family.Name:
                # check if the character after the root name is a none alphabetical character
                if len(family.Name) > len(duplicate_family_root_name):
                    if family.Name[len(duplicate_family_root_name)].isalpha() == False:
                        return family
    return None


def find_duplicate_families (doc):
    """
    Find duplicate families in the document.

    :param doc: Revit document
    :type doc: Autodesk.Revit.DB.Document
    :return: Dictionary of original family id and list of duplicate families.
    :rtype: dict[int, list[Autodesk.Revit.DB.Family]]
    """

    # dictionary in format source family_id: [duplicated family]
    duplicate_families = {}
    
    # get all families in the document
    all_families = get_category_name_to_family_dict(doc)
   
    for category_name, families in all_families.items():
        if len(families) > 1:
            # we have multiple families in this category, check for duplicates
            for family in families:
                duplicate_root_name = get_duplicate_family_root_name(family.Name)
                if duplicate_root_name:
                    # we have a duplicate family, find the original family
                    original_family = find_original_family_based_on_duplicate_family(all_families, family.Name, duplicate_root_name, category_name)
                    if original_family:
                        #print("Duplicate family found: {} (original: {})".format(family.Name, original_family.Name))
                        if original_family.Id not in duplicate_families:
                            duplicate_families[original_family.Id] = []
                        duplicate_families[original_family.Id].append(family)
                    else:
                        #print("Duplicate family found: {} (original not found)".format(family.Name))
                        pass
    
    return duplicate_families


def find_matching_types_between_families (source_family, target_family):
    """
    Finds matching types between two families based on type names.

    :param source_family: Source family to find matching types for.
    :type source_family: Autodesk.Revit.DB.Family

    :param target_family: Target family to find matching types in.
    :type target_family: Autodesk.Revit.DB.Family

    :return: Dictionary of source type id and target type id. If no match is found the source type id is not included in the dictionary.
    :rtype: dict[int, int]
    """

    matching_types = {}

    # get all types in the source family
    source_types = list(source_family.GetFamilySymbolIds())
    target_types = list(target_family.GetFamilySymbolIds())

    # compare type names
    for source_type_id in source_types:
        source_type = source_family.Document.GetElement(source_type_id)
        source_type_name = Element.Name.GetValue(source_type)
        for target_type_id in target_types:
            target_type = target_family.Document.GetElement(target_type_id)
            target_type_name = Element.Name.GetValue(target_type)
            if source_type_name == target_type_name:
                matching_types[source_type.Id.IntegerValue] = target_type.Id.IntegerValue
            else:
                # check if the type name starts with the source type name and is followed by " " and a number
                if target_type_name.startswith(source_type_name):
                    if len(target_type_name) > len(source_type_name) + 2:
                        if target_type_name[len(source_type_name)] == " " and target_type_name[len(source_type_name)+1].isnumeric():
                            matching_types[source_type.Id.IntegerValue] = target_type.Id.IntegerValue

    return matching_types


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions to load or reload families 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

clr.AddReference("System")
from System.Collections.Generic import List


# import common library

from duHast.Utilities import files_get as fileGet
from duHast.Utilities.Objects import result as res
from duHast.Utilities.files_io import was_file_edited_in_time_span, time_span_file_edited_last
from duHast.Revit.Family import family_utils as rFamUtil
from duHast.Revit.Family.family_load_option import *
from duHast.Revit.Family.family_reload_single import reload_family

from duHast.Revit.Family.Utility import loadable_family_categories as rFamLoadable
from duHast.UI.Objects.ProgressBase import ProgressBase

import Autodesk.Revit.DB as rdb

# --------------------------------------------------- Family Loading / inserting -----------------------------------------





def reload_all_families(doc, library_location, include_sub_folders=False, time_span_in_minutes=None, progress_callback=None, delete_new_types=True, report_matches_only=False):
    """
    Reloads a number of families with setting: parameter values overwritten: True


    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param library_location: _description_
    :type library_location: str
    :param include_sub_folders: _description_
    :type include_sub_folders: bool
    :param time_span_in_minutes: the span of time in which the file was modified to be considered for reload, defaults to None
    :type time_span_in_minutes: int, optional
    :param progress_callback: a progress call back function, defaults to None
    :type progress_callback: ProgressBase, optional
    :param delete_new_types: If true, any new family types introduced during the reload will be deleted from the project. Default is True.
    :type delete_new_types: bool
    :param report_matches_only: If true, will only report on families where a reload attempt was made, defaults to False
    :type report_matches_only: bool, optional
    
    :raises UserWarning: _description_

    :return: Returns True if any of the reload actions was successful.
    :rtype: bool
    """

    result = res.Result()
    
    try:

        # type checking
        if progress_callback != None:
            if not isinstance(progress_callback, ProgressBase):
                raise TypeError("progress_callback must be an instance of ProgressBase")
        
        # build library
        library = fileGet.files_as_dictionary(
            library_location, "", "", ".rfa", include_sub_folders
        )
        if len(library) == 0:
            result.update_sep(False, "Library is empty!")
            # get out...
            raise UserWarning("Empty Library")
        else:
            result.append_message("Found: {} families in Library!".format(len(library)))
        
        # get all families in file:
        family_ids = get_family_ids_from_symbols(doc)

        # check if anything to reload
        if len(family_ids)==0:
            message = "Found no loadable families in file!"
            result.update_sep(False, message)
            return result

        # update result
        result.append_message(
            "Found:  {} loadable families in file.".format(len(family_ids))
        )

        # set up progress
        progress_max = len(family_ids)
        callback_counter = 0

        # loop over families in file and attempt reload
        for fam_id in family_ids:

            # progress call back
            callback_counter += 1

            # get the family
            fam = doc.GetElement(fam_id)
            # get the family name
            fam_name = rdb.Element.Name.GetValue(fam)

            # check if family name exists in library
            if fam_name not in library:
                # set appropriate message
                message = "Family: {} not found in library.".format(fam_name)
                # check if progress reporting is desired
                if progress_callback != None:
                    # only report progress if not in report matches only mode
                    if not (report_matches_only):
                        progress_callback.update(callback_counter, progress_max, message)
                
                result.update_sep(
                    result.status, message
                )

                # skip to next family
                continue

            # proceed with reload
            # check if progress callback is set
            if progress_callback != None:
                # only report progress if not in report matches only mode
                if not (report_matches_only):
                    progress_callback.update(callback_counter, progress_max, "Found match for: {}".format(fam_name))

            result.append_message("Found match for: {}".format(fam_name))

            # check number of matches found
            if len(library[fam_name]) != 1:
                matches_message = ""
                for path in library[fam_name]:
                    matches_message = matches_message + "..." + path + "\n"
                matches_message = "Found multiple matches for {} \n {}".format(
                    fam_name, matches_message
                )
                matches_message = matches_message.strip()
                # found multiple matches for family by name only...aborting reload
                result.append_message(matches_message)

                # check if progress callback is set
                if progress_callback != None:
                    progress_callback.update(callback_counter, progress_max, matches_message)
                
                # skip to next family
                continue

            # found single match for family by name
            result.update_sep(
                True, "Found single match: {}".format(library[fam_name][0])
            )

            # check if family was changed within time span, if not skip reload
            if time_span_in_minutes is not None:
                if was_file_edited_in_time_span(library[fam_name][0], time_span_in_minutes) == False:
                    time_last_edited = time_span_file_edited_last(library[fam_name][0])
                    result.append_message("Family {} was not modified within time span of {} minutes. Edited last: {} ago...skipping reload.".format(fam_name, time_span_in_minutes, time_last_edited))
                    continue

            # attempt to load the family:
            load_result = reload_family(
                doc=doc, 
                family=fam, 
                family_file_path=library[fam_name][0], 
                delete_new_types=delete_new_types
            )

            # check results
            if load_result.status == True:
                # make sure that if a single reload was successful that this method returns true
                result.status = True

            # preserve reload log messages
            result.append_message(load_result.message)

            # return the reloaded family, if there is one
            if len(load_result.result) > 0:
                result.result.append(load_result.result[0])

            # check if progress callback is set
            if progress_callback != None:
                progress_callback.update(callback_counter, progress_max, load_result.message)

        
    except Exception as e:
        message = "Failed to load families with exception: {}".format(e)
        result.update_sep(False, message)
    return result


def get_family_ids_from_symbols(doc):
    """
    Get all loadable family ids in file

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids representing loadable families.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    family_ids = []
    # build list of all categories we want families to be reloaded of
    fam_cats = List[rdb.BuiltInCategory](rFamLoadable.CATEGORIES_LOADABLE_TAGS)
    fam_cats.AddRange(rFamLoadable.CATEGORIES_LOADABLE_TAGS_OTHER)
    fam_cats.AddRange(rFamLoadable.CATEGORIES_LOADABLE_3D)
    fam_cats.AddRange(rFamLoadable.CATEGORIES_LOADABLE_3D_OTHER)
    # get all symbols in file
    fam_symbols = rFamUtil.get_family_symbols(doc, fam_cats)
    # get families from symbols and filter out in place families
    for fam_symbol in fam_symbols:
        if (
            fam_symbol.Family.Id not in family_ids
            and fam_symbol.Family.IsInPlace == False
        ):
            family_ids.append(fam_symbol.Family.Id)
    return family_ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A base class used to store view filter rules.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


Stores view filters rules

The paraemeter id
the Evaluation type
the rule value


"""


#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#



import json

from duHast.Utilities.Objects import base

class ViewFilterRule(base.Base):
    def __init__(self, data_type="view filter rule", j=None, **kwargs):
        """
        Class constructor.
        """

        super(ViewFilterRule, self).__init__(**kwargs)

        self.data_type = data_type
        self.parameter_id = -1
        self.parameter_name = ""
        self.parameter_guid = ""
        self.evaluation_type = ""
        self.value_provider = ""
        self.rule_value = ""
        self.rule_type = ""
        self.is_inversed = False
        self.epsilon = None
       
        # check if any data was past in with constructor!
        if j is not None:
            # check type of data that came in:
            if type(j) == str:
                # a string
                j = json.loads(j)
            elif type(j) == dict:
                # no action required
                pass
            else:
                raise ValueError(
                    "Argument supplied must be of type string or type dictionary"
                )

            # load values and throw exception if something is missing!
            try:
                # get category ids
                self.parameter_id = j["parameter_id"]
                self.evaluation_type = j["evaluation_type"]
                self.rule_value = j["rule_value"]
                self.is_inversed = j["is_inversed"]
                self.parameter_name = j["parameter_name"]
                self.parameter_guid = j["parameter_guid"]
                self.rule_type = j["rule_type"]
                self.value_provider = j["value_provider"]
                self.epsilon = j.get("epsilon", None)

            except Exception as e:
                raise ValueError(
                    "Node {} failed to initialise with: {}".format(
                       data_type, e
                    )
                )
    

    def __eq__(self, other):
        """
        Override the default Equals behavior.

        Will take into account that shared parameters can have different ids in different documents

        """
        if isinstance(other, ViewFilterRule):
            # check if this is comparing a shared parameter based rule
            if self.parameter_guid != "" and other.parameter_guid != "":
                return self.parameter_guid == other.parameter_guid and \
                    self.evaluation_type == other.evaluation_type and \
                    self.rule_value == other.rule_value and \
                    self.is_inversed == other.is_inversed and \
                    self.parameter_name == other.parameter_name and \
                    self.rule_type == other.rule_type and \
                    self.value_provider == other.value_provider and \
                    self.epsilon == other.epsilon
            else:
                return self.parameter_id == other.parameter_id and \
                    self.evaluation_type == other.evaluation_type and \
                    self.rule_value == other.rule_value and \
                    self.is_inversed == other.is_inversed and \
                    self.parameter_name == other.parameter_name and \
                    self.parameter_guid == other.parameter_guid and \
                    self.rule_type == other.rule_type and \
                    self.value_provider == other.value_provider and \
                    self.epsilon == other.epsilon
        return False
    

    def __ne__(self, other):
        """
        Define a non-equality test
        """
        return not self.__eq__(other)
    

    def __hash__(self):
        """
        Override the default hash behavior (that returns the id or the object)
        """
        return hash((
            self.parameter_id, 
            self.evaluation_type, 
            self.rule_value, 
            self.is_inversed, 
            self.parameter_name, 
            self.parameter_guid,
            self.rule_type,
            self.value_provider,
            self.epsilon
            )
        )


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Room tag not in room warnings solver class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import result as res
from duHast.Revit.Rooms.room_tags import move_tag_to_room
from duHast.Revit.Common.transaction import in_transaction
from duHast.Revit.Warnings.warning_guids import ROOM_TAG_OUTSIDE_ROOM

from duHast.Utilities.Objects import base


class RevitWarningsSolverRoomTagToRoom(base.Base):
    def __init__(
        self,
        transaction_manager=in_transaction,
        callback=None,
    ):
        """
        Empty constructor this solver does not take any further arguments.
        """

        # ini super class to allow multi inheritance in children!
        super(RevitWarningsSolverRoomTagToRoom, self).__init__()
        self.filter_name = "Room tag outside of room."
        self.transaction_manager = transaction_manager
        self.callback = callback

    # --------------------------- room tag not in room ---------------------------
    #: guid identifying this specific warning
    GUID = ROOM_TAG_OUTSIDE_ROOM

    def solve_warnings(self, doc, warnings):
        """
        Solver moving room tags to room location point.

        :param doc: Current Revit model document.
        :type doc: Autodesk.Revit.DB.Document
        :param warnings: List of warnings to be solved.
        :type warnings: Autodesk.Revit.DB.FailureMessage

        :return:
            Result class instance.

            - .status True if all room tags within warnings could be moved to room location point. Otherwise False.
            - .message will be 'moved tag to room xyz'

        :rtype: :class:`.Result`
        """

        return_value = res.Result()
        if len(warnings) > 0:

            # set up progress counter
            counter = 1

            for warning in warnings:
                # report progress to call back if required
                if self.callback:
                    self.callback.update(counter, len(warnings))

                element_ids = warning.GetFailingElements()
                for el_id in element_ids:
                    result = move_tag_to_room(
                        doc=doc,
                        tag_id=el_id,
                        transaction_manager=self.transaction_manager,
                    )
                    return_value.update(result)

                # increase progress counter
                counter = counter + 1

                # check if cancelled
                if self.callback:
                    if self.callback.is_cancelled():
                        return_value.append_message("User cancelled!")
                        break
        else:
            return_value.update_sep(
                True,
                "{}: No warnings of type: room tag outside of room in model.".format(
                    self.filter_name
                ),
            )
        return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit building pads helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Common.common import get_ids_from_element_collector
from duHast.Revit.BuildingPads.Utility.RevitBuildingPadsFilter import (
    _get_all_building_pad_types_by_category,
    _get_building_pad_types_by_class,
)

# import Autodesk
# Guard buildingpad import ( no longer supported in Revit 2024 and up )
try:
    from Autodesk.Revit.DB import BuildingPad as BuildingPad
    HAS_BUILDINGPAD = True
except Exception:
    HAS_BUILDINGPAD = False
    BuildingPad = None


from Autodesk.Revit.DB import BuiltInCategory, FilteredElementCollector

# --------------------------------------------- utility functions ------------------


def get_all_building_pad_types_by_category(doc):
    """
    Gets a filtered element collector of all BuildingPad types in the model.

    - Basic BuildingPad

    Filters by category.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing building pad types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = _get_all_building_pad_types_by_category(doc)
    return collector


def get_building_pad_types_by_class(doc):
    """
    Gets a filtered element collector of all building pad types in the model:

    - Basic BuildingPad

    Filters by class.
    Since there are no in place families of type building pad possible, this should return the same elements as the by category filter.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing building pad types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = _get_building_pad_types_by_class(doc)
    return collector


# -------------------------------- none in place BuildingPad types -------------------------------------------------------


def get_all_building_pad_instances_in_model_by_category(doc):
    """
    Gets all building pad elements placed in model.

    Filters by category.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing ceiling instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    if not HAS_BUILDINGPAD:
        raise EnvironmentError("BuildingPad class is not available.")
    
    return (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_BuildingPad)
        .WhereElementIsNotElementType()
    )


def get_all_building_pad_instances_in_model_by_class(doc):
    """
    Gets all building pad elements placed in model.

    Filters by class.
    Since there are no in place families of type building pad possible, this should return the same elements as the by category filter.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing ceiling instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    if not HAS_BUILDINGPAD:
        raise EnvironmentError("BuildingPad class is not available.")
    return (
        FilteredElementCollector(doc)
        .OfClass(BuildingPad)
        .WhereElementIsNotElementType()
    )


def get_all_building_pad_type_ids_in_model_by_category(doc):
    """
    Gets all building pad element type ids available in model.

    Filters by category.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing building pad type ids.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    ids = []
    col_cat = get_all_building_pad_types_by_category(doc)
    ids = get_ids_from_element_collector(col_cat)
    return ids


def get_all_building_pad_type_ids_in_model_by_class(doc):
    """
    Gets all building pad element type ids available in model.

    Filters by class.
    Since there are no in place families of type building pad possible, this should return the same elements as the by category filter.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing building pad type ids.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    ids = []
    col_class = get_building_pad_types_by_class(doc)
    ids = get_ids_from_element_collector(col_class)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit view filters to storage conversion. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
clr.AddReference('System')
from System.Collections.Generic import List
from System import Enum

from duHast.Utilities.Objects.result import Result

from duHast.Revit.Views.Objects.Data.view_filter import ViewFilter
from duHast.Revit.Views.Objects.Data.view_filter_rule import ViewFilterRule
from duHast.Revit.Views.Objects.Data.view_filter_logic_container import ViewFilterLogicContainer
from duHast.Revit.SharedParameters.shared_parameters import get_all_shared_parameters
from duHast.Revit.Common.parameter_project import get_project_parameter_definitions

from Autodesk.Revit.DB import (
    BuiltInParameter,
    ElementId,
    Element, 
    LogicalAndFilter, 
    ElementFilter, 
    ElementParameterFilter,
    FilteredElementCollector,
    FilterElementIdRule, 
    FilterNumericValueRule, 
    FilterInverseRule, 
    FilterStringRule,
    LogicalOrFilter,  
    ParameterValueProvider,
)


DEBUG = False


# Note:
# Inverse rules are essentially wrappers around standard rules inversing that outcome

def get_project_parameters(doc):
    """
    Get all project parameters in the document.

    :param doc: Current Revit document
    :type doc: Autodesk.Revit.DB.Document
    :return: Dictionary where key is the id of the parameter and values are a list of [name, guid]
    :rtype: dict
    """

    # get project parameter definitions
    definitions = get_project_parameter_definitions(doc)

    # get all shared parameters
    shared_parameters = get_all_shared_parameters(doc)


    # build dictionary where key is the id of the parameter and values are a list of [name, guid]
    # guid will be None for non-shared parameters
    project_parameters = {}

    # loop over definitions and find any shared parameters
    for d in definitions:

        # set default flag to not shared
        is_shared = False

        # loop over shared parameters to see if we have a match
        for sp in shared_parameters:
            if d.Id.IntegerValue == sp.Id.IntegerValue:
                # this is a shared parameter
                is_shared = True
                project_parameters[d.Id.IntegerValue] = [d.Name, str(sp.GuidValue)]
                break
        
        # this is not a shared parameter
        if not is_shared:
            project_parameters[d.Id.IntegerValue] = [d.Name, None]
    
    return project_parameters


def analyze_rule(doc, rule,  is_inversed, project_parameters, nesting_level, debug = False):
    """
    
    Analyze a single rule and return a view filter rule object.
    
    :param doc: The Revit document.
    :type doc: Document
    :param rule: The rule to analyze.
    :type rule: FilterRule
    :param is_inversed: Whether the rule is inversed.
    :type is_inversed: bool
    :param nesting_level: The nesting level of the rule.
    :type nesting_level: int
    :param debug: Whether to print debug information.
    :type debug: bool

    :return: A result object containing the view filter rule. If an error occurs, the result object will indicate failure and contain an error message.
    :rtype: :class:`.Result`
    """

    return_value = Result()
    try:

        view_filter_rule = ViewFilterRule()

        # can be numeric or a string rule
        # get the parameter to be checked
        if debug:
            return_value.append_message ("{} rule parameter id: {}".format("..." * nesting_level, rule.GetRuleParameter().IntegerValue))
        
        # parameter id
        view_filter_rule.parameter_id = rule.GetRuleParameter().IntegerValue

        # save the rule type
        view_filter_rule.rule_type = type(rule).__name__

        # depending on the rule type there might be an epsilon value ( in the moment only for double values)
        if hasattr(rule, "Epsilon"):
            view_filter_rule.epsilon = rule.Epsilon
            if debug:
                return_value.append_message ("{} rule epsilon: {}".format("..." * nesting_level, view_filter_rule.epsilon))


        # in the moment there is only one value provider type
        view_filter_rule.value_provider = ParameterValueProvider.__name__

        # if the id is negative means its a built in parameter and we dont have to get its name and or guid
        # check if parameter id indicates a custom parameter by checking if id value is greater than 0
        # and whether we have any project parameters to check against
        if view_filter_rule.parameter_id > 0 and len(project_parameters) > 0:
            
            if view_filter_rule.parameter_id in project_parameters:
               
                # assign the parameter name
                view_filter_rule.parameter_name = project_parameters[view_filter_rule.parameter_id][0]
                
                if debug:
                    return_value.append_message ("{} rule parameter name: {}".format("..." * nesting_level, view_filter_rule.parameter_name))
                
                if project_parameters[view_filter_rule.parameter_id][1] is not None:
                    # assign the parameter guid
                    view_filter_rule.parameter_guid = project_parameters[view_filter_rule.parameter_id][1]
                    
                    if debug:
                        return_value.append_message ("{} rule parameter guid: {}".format("..." * nesting_level, view_filter_rule.parameter_guid))
                else:
                    # guid is None, leave the default value
                    if debug:
                        return_value.append_message ("{} rule parameter guid: None".format("..." * nesting_level))
                    
        elif view_filter_rule.parameter_id < 0:
                # built in parameter, leave the default name and guid values
                if debug:
                    return_value.append_message ("{} rule parameter is built in parameter".format("..." * nesting_level))
                
                # get the built in parameter name
                param = Enum.ToObject(BuiltInParameter, view_filter_rule.parameter_id)
                view_filter_rule.parameter_name = str(param)
        else:
            pass
            # should never get here but just in case

        # get the evaluation type (ends with, starts with, equals, greater than, etc)
        if debug:
            return_value.append_message ("{} rule evaluator: {}".format("..." * nesting_level, rule.GetEvaluator().GetType().Name))
        
        view_filter_rule.evaluation_type = rule.GetEvaluator().GetType().Name

        # get the rule value
        if isinstance(rule, FilterElementIdRule):
            # check for element id rule
            if debug:
                return_value.append_message ("{} rule value: {}".format("..." * nesting_level, rule.RuleValue.IntegerValue))
            view_filter_rule.rule_value=rule.RuleValue.IntegerValue
        
        elif isinstance(rule, FilterNumericValueRule):
            # check for numeric rule
            if debug:
                return_value.append_message ("{} rule value: {}".format("..." * nesting_level, rule.RuleValue))
            
            # check if the rule value is an element id ( if a global parameter provides the value, the rule value is the global parameter id)
            if isinstance(rule.RuleValue, ElementId):
                view_filter_rule.rule_value=rule.RuleValue.IntegerValue
            else:
                view_filter_rule.rule_value=rule.RuleValue
        
        elif isinstance(rule, FilterStringRule):
            # check for string rule
            if debug:
                return_value.append_message ("{} rule value: {}".format("..." * nesting_level, rule.RuleString))
            view_filter_rule.rule_value=rule.RuleString
        
        # is the rule inversed
        if debug:
            return_value.append_message ("{} is inversed: {}".format("..." * nesting_level, is_inversed))
        
        view_filter_rule.is_inversed = is_inversed

        if debug:
            return_value.append_message ("{} rule complete: {}".format("..." * nesting_level, view_filter_rule))
        
        return_value.result.append(view_filter_rule)
        return return_value
    except Exception as e:
        return_value.update_sep(False, "Failed to analyze rule. Error: {}".format(e))
        return return_value

    

def analyze_element_parameter_filter(doc, element_parameter_filter, project_parameters, nesting_level, debug = False):
    """
    Analyze an element parameter filter and return a list of view filter rules.

    :param doc: The Revit document.
    :type doc: Document
    :param element_parameter_filter: The element parameter filter to analyze.
    :type element_parameter_filter: ElementParameterFilter
    :param nesting_level: The nesting level of the filter.
    :type nesting_level: int
    :param debug: Whether to print debug information.
    :type debug: bool

    :return: A result object containing a list of view filter rules. If an error occurs, the result object will indicate failure and contain an error message.
    :rtype: :class:`.Result`
    """

    return_value = Result()

    try:
        rules = element_parameter_filter.GetRules()
        if debug:
            return_value.append_message ("{} rules: {}".format("..." * nesting_level, rules.Count))
        
        # set up a list to hold the rules
        rules_analysed = []

        # go over all rules
        for rule in rules:
            if debug:
                return_value.append_message ("{} rule type: {}".format("..." * nesting_level, type(rule)))
        
            # check if the rule is an inverse rule
            if isinstance(rule, FilterInverseRule):
                if debug:
                    return_value.append_message ("{} is inverse rule...unwrap".format("..." * nesting_level))
                
                # get the nested rule
                rule_nested = rule.GetInnerRule()

                # analyze the inverse rule
                rule_analysed_result = analyze_rule(doc, rule_nested, True, project_parameters, nesting_level+1, debug)

                # check if successful
                if not rule_analysed_result.status:
                    return_value.update_sep(False, "Failed to analyze element parameter filter inverse rule. Error: {}".format(rule_analysed_result.message))
                    continue
                
                # add generated rule to list
                rules_analysed.append(rule_analysed_result.result[0])
            else:
                # analyze the rule
                rule_analysed_result = analyze_rule(doc, rule, False, project_parameters,nesting_level+1, debug)

                # check if successful
                if not rule_analysed_result.status:
                    return_value.update_sep(False, "Failed to analyze element parameter filter rule. Error: {}".format(rule_analysed_result.message))
                    continue
                
                # add generated rule to list
                rules_analysed.append(rule_analysed_result.result[0])
        

        # return the rules
        return_value.result.append(rules_analysed)
        return return_value

    except Exception as e:
        return_value.update_sep(False, "Failed to analyze element parameter filter. Error: {}".format(e))
        return return_value
       


def analyze_logical_filter(doc, logical_filter, project_parameters, nesting_level=0, debug = False):
    """
    Analyze a logical filter and return a view filter logic container.

    :param doc: The Revit document.
    :type doc: Document
    :param logical_filter: The logical filter to analyze.
    :type logical_filter: LogicalAndFilter or LogicalOrFilter
    :param nesting_level: The nesting level of the filter.
    :type nesting_level: int
    :param debug: Whether to print debug information.
    :type debug: bool

    :return: A result object containing a view filter logic container. If an error occurs, the result object will indicate failure and contain an error message.
    :rtype: :class:`.Result`
    """

    return_value = Result()

    try:
        # setup a logical container
        logical_container = ViewFilterLogicContainer()

        # store the type of logical container
        logical_container.logic_container_type = type(logical_filter).__name__

        # get the filters in the logical filter
        # should always be a list of element parameter filters or nested logical filters
        filters = logical_filter.GetFilters()

        if debug:
            return_value.append_message ("{} Logical filter contains the following filters:".format("..." * nesting_level))
            return_value.append_message ("{} {} ".format("..." * nesting_level, type(filters)))

        # set up a comparing type to make sure we get a list
        filter_list_type = List[ElementFilter]

        # should always be a list of element parameter filters or nested logical filters
        if isinstance(filters, filter_list_type):

            # loop over list contents
            for filter in filters:
                # check the type of filter
                if isinstance(filter, ElementParameterFilter):
                    if debug:
                        return_value.append_message ( "{} is element parameter filter".format("..." * nesting_level))
                    
                    # analyze the element parameter filter
                    rules_result = analyze_element_parameter_filter(doc, filter, project_parameters, nesting_level + 1, debug)
                    
                    # check what came back
                    if rules_result.status and len(rules_result.result) > 0:
                        if debug:
                            return_value.append_message ("{} adding {} rules to logical container".format("..." * nesting_level, len(rules_result.result)))
                        
                        # add rules to the logical container
                        logical_container.view_filter_rules = logical_container.view_filter_rules + rules_result.result[0]
                    else:
                        # something went wrong
                        return_value.update_sep(False, "Failed to analyze element parameter filter. Error: {}".format(rules_result.message))

                elif isinstance(filter, LogicalAndFilter) or isinstance(filter, LogicalOrFilter):
                    if debug:
                        return_value.append_message ( "{} is logical and filter...recursive call".format("..." * nesting_level))
                    
                    # analyse another logical conditions
                    nested_container_result = analyze_logical_filter(doc,filter, project_parameters, nesting_level + 1, debug)
                    
                    # check what came back
                    if  nested_container_result.status and len(nested_container_result.result) > 0:
                        # add nested container to the logical container
                        logical_container.logic_containers.append(nested_container_result.result[0])
                        
        else:
            # not sure what this...
            return_value.update_sep(False, "{} Currently not supported: {}".format("..." * nesting_level, type(filters)))
            return return_value
        
        if debug:
            return_value.append_message ("{} logical filter complete: {}".format("..." * nesting_level, logical_container))
        
        # add the container
        return_value.result.append(logical_container)

        # return to caller
        return return_value

    except Exception as e:
        return_value.update_sep(False, "Failed to analyze logical filter. Error: {}".format(e))
        return return_value



def analyze_filters(doc, filters,  progress_callback, debug = False):
    """
    Analyze all view filters in the document and return a list of view filter objects.

    :param doc: The Revit document.
    :type doc: Document
    :param filters: The view filters to analyze.
    :type filters: FilterElementCollector
    :param forms: The pyrevit forms module.
    :type forms: module
    :param debug: Whether to print debug information.
    :type debug: bool

    :return: A result object containing a list of view filter objects. If an error occurs, the result object will indicate failure and contain an error message.
    :rtype: :class:`.Result`
    """

    return_value = Result()

    try:

        analysed_filters = []

        # get all project parameters in order to be able to identify any shared parameters properly when re-importing the filters
        project_parameters = get_project_parameters(doc)

        # prepare progress bar
        max_value = 0

        # check if we got a FilteredElementCollector or a list
        if isinstance(filters,FilteredElementCollector):
            max_value = len(filters.ToElements())
        else:
            max_value = len(filters)
        
        counter = 1

        # loop over view filters in the model
        for filter in filters:
            
            view_filter = ViewFilter()

            # store the filter name
            view_filter.name = Element.Name.GetValue(filter)

            if debug:
                # get the filter name
                return_value.append_message ("filter name: {}".format(Element.Name.GetValue(filter)))

            # update progress bar
            if progress_callback:
                progress_callback.update(counter, max_value, view_filter.name)
           
            # getting the revit category ids the filter is applied to
            filter_revit_category_ids = filter.GetCategories()
            
            # add category ids to the view filter
            for id in filter_revit_category_ids:
                if debug:
                    return_value.append_message ("...Filter Id [{}]".format(id.IntegerValue))
                
                view_filter.category_ids.append(id.IntegerValue)
            
            # getting the filter elements
            filter_elements = filter.GetElementFilter()

            # check the type of filter, should be a logical element filter (top level)
            if isinstance(filter_elements, LogicalAndFilter) or isinstance(filter_elements, LogicalOrFilter):
                if debug:
                    return_value.append_message ( "...is logical filter")
                
                # analyze the logical filter
                container_host_result = analyze_logical_filter(doc, filter_elements,  project_parameters,   1, debug)
                
                # check what came back
                if container_host_result.status and len(container_host_result.result) > 0:

                    # store the container host in the view filter
                    view_filter.logic_containers.append(container_host_result.result[0])
                    
                    if debug:
                        return_value.append_message ("...updated container host to view filter")
                    
                    # add the view filter to the list to be returned
                    analysed_filters.append(view_filter)
            else:

                # not sure what this...
                return_value.append_message ("    Currently not supported: {}".format(type(filter_elements)))

            # update progress
            counter = counter + 1

            # check for user cancel
            if progress_callback != None:
                if progress_callback.is_cancelled():
                    return_value.append_message("User cancelled!")
                    break
        
        return_value.result.append(analysed_filters)
        return return_value

    except Exception as e:
        return_value.update_sep(False, "Failed to analyze filters. Error: {}".format(e))
        return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit view filters storage creation from json. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import clr
clr.AddReference('System')
from System.Collections.Generic import List


from duHast.Utilities.Objects.result import Result

from duHast.Revit.Views.Utility.convert_data_to_filter_rule import convert_data_to_rule
from duHast.Revit.Views.Utility.convert_data_to_filter_logic_filter import get_logical_filter_class
from duHast.Revit.Views.filters import create_filter, get_all_filters, update_filter


from Autodesk.Revit.DB import (
    Category,
    ElementId,
    ElementFilter,
    ElementParameterFilter,
    FilterRule,
)

DEBUG = False

def import_rules_from_data(doc, data_rules):

    # set up a status tracker
    return_value = Result()

    # container for new rules
    element_parameter_filters = List[ElementFilter]()

    # loop over filters in logic container and convert to revit filter rules
    for filter in  data_rules.view_filter_rules:

        return_value.append_message("...Importing filter: {}".format(filter.parameter_name))

        conversion_result = convert_data_to_rule(doc, filter)
        
        if conversion_result.status:
            return_value.append_message("...Successfully converted filter: {}".format(filter.parameter_name))
            
            # rules need to be added to an ElementParameterFilter object which in turns get added to the logic container
            # convert python list to c# list
            rules_net =  List[FilterRule]()

            for rule in conversion_result.result:
                rules_net.Add(rule)
            
            # create the element parameter filter using rules provided
            element_parameter_filter = ElementParameterFilter(rules_net)

            # add to container
            element_parameter_filters.Add(element_parameter_filter)
            
        else:
            return_value.update_sep(False, "...Failed to convert filter: {}. Error: {}".format(filter.parameter_name, conversion_result.message))


    return_value.result.append(element_parameter_filters)
    return return_value


def import_nested_logic_container_from_data(doc, logic_filter_containers):

    # set up a status tracker
    return_value = Result()
    
    try:
        # container for nested container
        element_parameter_filters = List[ElementFilter]()

        for logic_filter_container in  logic_filter_containers:

            # container for nested container
            nested_element_parameter_filters = List[ElementFilter]()

            # get the logic container filter class
            logic_filter_class = get_logical_filter_class(logic_filter_container.logic_container_type)
            
            if not logic_filter_class:
                return_value.update_sep(False, "Failed to get logic filter class for type: {}.".format(logic_filter_container.logic_container_type))
                return return_value
            
            return_value.append_message("Using logic filter class: {} for logic container of type: {}".format(logic_filter_class.__name__, logic_filter_container.logic_container_type))
            
            if DEBUG:
                print("Using logic filter class: {} for logic container of type: {}".format(logic_filter_class.__name__, logic_filter_container.logic_container_type))

            # get the nested logic containers 
            nested_logic_containers = logic_filter_container.logic_containers
            
            # check if there are any nested containers
            if nested_logic_containers is None or len(nested_logic_containers) == 0:
                return_value.append_message("No logic container found. Skipping logic containers...")
            else:
                # recursive call
                import_nested_containers_result = import_nested_logic_container_from_data(doc, nested_logic_containers)

                if not import_nested_containers_result.status:
                    return_value.update_sep(False, "Failed to import nested logic container. Error: {}".format(import_nested_containers_result.message))
                    # check if any rules were imported
                    if len(import_nested_containers_result.result) > 0:
                        for entry in import_nested_containers_result.result[0]:
                            nested_element_parameter_filters.Add(entry)

                else:
                    return_value.append_message("Successfully imported {} rules for nested logic container.".format(len(import_nested_containers_result.result[0])))
                    # add the imported rules to the main container
                    for entry in import_nested_containers_result.result[0]:
                        nested_element_parameter_filters.Add(entry)

            # look at view filter rules at this level
            if len(logic_filter_container.view_filter_rules) == 0:
                return_value.append_message("No filter rules found. Skipping view filter rules...")
                
                if DEBUG:
                    print ("no rules\n")
            else:

                if DEBUG:
                    print ("got rules {}\n".format(len(logic_filter_container.view_filter_rules)))

                # import the rules
                import_rules_result = import_rules_from_data(doc, logic_filter_container)

                # user feedback
                if not import_rules_result.status:
                    return_value.update_sep(False, "Failed to import rules for view filter. Error: {}".format(import_rules_result.message))
                    if DEBUG:
                        print("Failed to import rules for view filter. Error: {}\n".format(import_rules_result.message))
                else:
                    return_value.append_message("Successfully imported {} rules for view filter".format(len(import_rules_result.result[0])))
                    if DEBUG:
                        print("Successfully imported {} rules for view filter\n".format(len(import_rules_result.result[0])))

                
                #element_parameter_filters = List[ElementFilter]()

                # transfer rules across
                for entry in import_rules_result.result[0]:
                    if DEBUG:
                        print("adding rule {}\n".format(entry))
                    nested_element_parameter_filters.Add(entry)
                
                # user feedback
                if DEBUG:
                    print("imported rules {}\n".format(nested_element_parameter_filters.Count))


            try:
                # create the logic container class instance
                logic_container_class_instance = logic_filter_class(nested_element_parameter_filters)
                element_parameter_filters.Add(logic_container_class_instance)
                return_value.append_message("Successfully created logic container of type: {}".format(logic_filter_container.logic_container_type))
                if DEBUG:
                    print("Successfully created logic container of type: {}\n".format(logic_filter_container.logic_container_type))
            except Exception as e:
                return_value.update_sep(False, "Failed to create logic container of type: {}. Error: {}".format(logic_filter_container.logic_container_type, e))
                if DEBUG:
                    print("Failed to create logic container of type: {}. Error: {}\n".format(logic_filter_container.logic_container_type, e))

        
        # add the overall list to the return value
        return_value.result.append(element_parameter_filters)
    except Exception as e:
        return_value.update_sep(False, "Failed to import nested logic container. Error: {}".format(e))

    return return_value
    


def import_root_logic_container_from_data(doc, view_filter_json):

    # set up a status tracker
    return_value = Result()
    
    try:
        return_value.append_message("Importing root logic container for view filter: {}".format(view_filter_json.name))
        
        # get the logic containers at the top of the tree
        logic_filter_container = view_filter_json.logic_containers
        if logic_filter_container is None or len(logic_filter_container) == 0:
            return_value.append_message("No logic container found. Skipping view filter: {}".format(view_filter_json.name))
            if DEBUG:
                print("No logic container found. Skipping view filter: {}\n".format(view_filter_json.name))
            return return_value

        return_value.append_message("Found {} root logic container(s) for view filter: {}".format(len(logic_filter_container), view_filter_json.name))
        # there is no need to check for rules here as this is done in the nested logic container function ( root container has no rules at the same level)
        if DEBUG:
            print("Found {} root logic container(s) for view filter: {}\n".format(len(logic_filter_container), view_filter_json.name))

        # get the root container json object (there is always only just one root container)
        logic_filter_root_container = logic_filter_container[0]

        # get the logic filter class
        logic_container_class = get_logical_filter_class(logic_filter_root_container.logic_container_type)
        if not logic_container_class:
            return_value.update_sep(False, "Failed to get logic filter class for type: {}. Skipping view filter: {}".format(logic_filter_root_container.logic_container_type, view_filter_json.name))
            return return_value
    
        return_value.append_message("Using logic container class: {} for view filter: {}".format(logic_container_class.__name__, view_filter_json.name))
        if DEBUG:
            print("Using logic container class: {} for view filter: {}\n".format(logic_container_class.__name__, view_filter_json.name))

        # container for nested container(s)
        element_parameter_filters = List[ElementFilter]()

        # check if there are any nested containers to import
        if len(logic_filter_root_container.logic_containers)> 0:
            # got nested containers
            if DEBUG:
                print("Importing root logic containers: {} of type: {}".format(len( logic_filter_root_container.logic_containers), logic_filter_root_container.logic_container_type))

            # import the nested logic containers (there might be none if all we have are rules within the root logic container)
            import_nested_logic_result = import_nested_logic_container_from_data(doc, logic_filter_root_container.logic_containers)

            if DEBUG:
                print("here {}\n".format(view_filter_json.name),import_nested_logic_result.result)

            # check what we got...
            if not import_nested_logic_result.status:
                # only rules within the root container
                return_value.append_message("No nested logic containers found for {}".format(view_filter_json.name))
                # check if any rules were imported ( .net list!!)
                if import_nested_logic_result.result[0].Count>0:
                    return_value.append_message("Successfully imported {} rules for root logic container for view filter: {}".format(import_nested_logic_result.result[0].Count, view_filter_json.name))
                    for entry in import_nested_logic_result.result[0]:
                        element_parameter_filters.Add(entry)
                else:
                    return_value.append_message("No rules found for root container for view filter: {}".format(view_filter_json.name))
            else:
                # nested containers and/or rules found
                return_value.append_message("Successfully imported {} rules for nested logic container for view filter: {}".format(len(import_nested_logic_result.result[0]), view_filter_json.name))
                # add the imported rules to the main container
                for entry in import_nested_logic_result.result[0]:
                    element_parameter_filters.Add(entry)
        else:
            # no nested containers...need to check for rules at this level
            if DEBUG:
                print("No nested logic containers found for {}\n".format(view_filter_json.name))
            return_value.append_message("No nested logic containers found for {}".format(view_filter_json.name))

            if DEBUG:
                print("Checking for rules at root level for view filter: {}\n".format(view_filter_json.name))
                print("Number of rules at root level: {}\n".format(len(logic_filter_root_container.view_filter_rules)))

            if len(logic_filter_root_container.view_filter_rules) == 0:
                return_value.append_message("No filter rules found. Skipping view filter rules for view filter: {}".format(view_filter_json.name))
                if DEBUG:
                    print ("no rules\n")
                return return_value
            
            try:
                rules_result = import_rules_from_data(doc, logic_filter_root_container)

                if DEBUG:
                    print("here2 {}\n".format(view_filter_json.name),rules_result.result)

                # add filters to main container
                for entry in rules_result.result[0]:
                    element_parameter_filters.Add(entry)

            except Exception as e:
                return_value.update_sep(False, "Failed to import rules at root level for view filter: {}. Error: {}".format(view_filter_json.name, e))
                if DEBUG:
                    print("Failed to import rules at root level for view filter: {}. Error: {}\n".format(view_filter_json.name, e))
                return return_value
        
        
        # add the imported rules to the return value
        logic_container_class_instance = logic_container_class(element_parameter_filters)

        return_value.result.append(logic_container_class_instance)
    except Exception as e:
        return_value.update_sep(False, "Failed to import root logic container for view filter: {}. Error: {}".format(view_filter_json.name, e))
    
    return return_value


def create_filter_from_json(doc, view_filter_json, element_filters):

    return_value = Result()
    try:
        # set up a .net list of Element Ids
        category_ids = List[ElementId]()

        # loop over category ids and add as element id
        for category_id in view_filter_json.category_ids:
            category_ids.Add(ElementId(category_id))
            
        # check we got a least one category
        if category_ids.Count == 0:
            return_value.update_sep(False, "No valid categories found for filter: {}.".format(view_filter_json.name))
            return return_value

        # create the filter
        filter_element = create_filter(
            doc, 
            view_filter_json.name,
            category_ids,
            element_filters,
        )

        # check result
        if not filter_element:
            return_value.update_sep(False, "Failed to create filter: {}.".format(view_filter_json.name))
            return return_value

        # add success message
        return_value.append_message("Successfully created filter: {}".format(view_filter_json.name))

        # return created filter
        return_value.result.append(filter_element)
    except Exception as e:
        
        return_value.update_sep(False, "Failed to create filter: {}. Error: {}".format(view_filter_json.name, e))
    return return_value


def update_filter_from_json(doc, existing_filter, view_filter_json, element_filters):

    return_value = Result()
    try:
        # set up a .net list of Element Ids
        category_ids = List[ElementId]()

        # loop over category ids and add as element id
        for category_id in view_filter_json.category_ids:
            category_ids.Add(ElementId(category_id))
            
        # check we got a least one category
        if category_ids.Count == 0:
            return_value.update_sep(False, "No valid categories found for filter: {}.".format(view_filter_json.name))
            return return_value

        # update the filter
        updated_filter_element = create_filter(
            doc, 
            view_filter_json.name,
            category_ids,
            element_filters,
            existing_filter,
        )

        # check result
        if not updated_filter_element:
            return_value.update_sep(False, "Failed to update filter: {}.".format(view_filter_json.name))
            return return_value

        # add success message
        return_value.append_message("Successfully updated filter: {}".format(view_filter_json.name))

        # return created filter
        return_value.result.append(updated_filter_element)
    except Exception as e:
        
        return_value.update_sep(False, "Failed to update filter: {}. Error: {}".format(view_filter_json.name, e))
    return return_value


def get_filters_in_model_by_name(doc):
    """
    Get all view filters in the model by name.

    :param doc: Revit Document
    :return: Dictionary of view filters by name
    """
    filters_in_model = get_all_filters(doc)
    filters_by_name = {f.Name: f for f in filters_in_model}
    return filters_by_name


def import_view_filters_from_data(doc, json_object, progress_callback=None, overwrite_existing=False):

    # set up a status tracker
    return_value = Result()

    counter = 0
    max_value = len(json_object)

    # get all filter in the model to check if we need to overwrite any
    existing_filters_by_name = get_filters_in_model_by_name(doc)

    # loop over json objects and create view filters
    for view_filter_json in json_object:

        # check if we have any existing filters
        is_existing_filter = view_filter_json.name in existing_filters_by_name

        #TODO: check if filter already exists and handle overwrite_existing flag
        if not overwrite_existing and is_existing_filter:
            return_value.append_message("Skipping existing view filter: {}".format(view_filter_json.name))
            if DEBUG:
                print("Skipping existing view filter: {}\n".format(view_filter_json.name))
            # update progress
            counter = counter + 1
            if progress_callback:
                progress_callback.update(counter, max_value, view_filter_json.name)
            continue

        if progress_callback:
            progress_callback.update(counter, max_value, view_filter_json.name)

        return_value.append_message("Importing view filter: {}".format(view_filter_json.name))

        if DEBUG:
            print("\nImporting view filter: {}\n".format(view_filter_json.name))
        
        try:
            # get the container and all its nested items
            container_result = import_root_logic_container_from_data(doc, view_filter_json)

            if not container_result.status:
                return_value.update_sep(False, "Failed to import logic container for view filter: {}. Error: {}".format(view_filter_json.name, container_result.message))
                continue

            # keep track of the container
            return_value.append_message("Successfully imported logic container for view filter: {}".format(view_filter_json.name))

            # set up a Revit view filter depending on whether this is a new filter or an update of an existing one
            if is_existing_filter:
                return_value.append_message("Updating existing filter: {}".format(view_filter_json.name))
                if DEBUG:
                    print("Updating existing filter: {}\n".format(view_filter_json.name))
                create_filter_result = update_filter_from_json(doc, existing_filters_by_name[view_filter_json.name], view_filter_json, container_result.result[0])
            else:
                return_value.append_message("Creating new filter: {}".format(view_filter_json.name))
                if DEBUG:
                    print("Creating new filter: {}\n".format(view_filter_json.name))
                create_filter_result = create_filter_from_json(doc, view_filter_json, container_result.result[0])

            # check result
            if not create_filter_result.status:
                return_value.update_sep(False, "Failed to create filter for view filter: {}. Error: {}".format(view_filter_json.name, create_filter_result.message))
                continue

            # create success message
            return_value.append_message("Successfully created filter for view filter: {}".format(view_filter_json.name))

            # return created filter
            return_value.result.append(create_filter_result.result[0])

        except Exception as e:
            return_value.update_sep(False, "Failed to import view filter: {}. Error: {}".format(view_filter_json.name, e))
            continue


        # update progress
        counter = counter + 1

        # check for user cancel
        if progress_callback != None:
            if progress_callback.is_cancelled():
                return_value.append_message("User cancelled!")
                break
    
    return return_value
                   


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains the view property utilities.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


def convert_view_data_to_list(view_data, headers):
    """
    Converts a list of dictionaries of view properties names and values to a list of properties only.

    :param view_data: List of dictionaries representing view properties
    :type view_data: [{}]
    :param headers: list of properties
    :type headers: [str]
    :return: A list of lists of view property values.
    :rtype: [[str]]
    """

    data = []
    for view_d in view_data:
        data_row = []
        for header in headers:
            if header in view_d:
                data_row.append(view_d[header])
            else:
                data_row.append("Property does not exist on element.")
        data.append(data_row)
    return data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to purging Revit ceilings. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family import purge_unused_family_types as rFamPurge
from duHast.Revit.Common import purge_utils as rPurgeUtils
from duHast.Revit.Ceilings import ceilings as rCeiling
from duHast.Revit.Ceilings.Utility import ceilings_type_sorting as rCeilingTypeSort


#: Built in family name for compound ceilings
COMPOUND_CEILING_FAMILY_NAME = "Compound Ceiling"

#: Built in family name for basic ceilings
BASIC_CEILING_FAMILY_NAME = "Basic Ceiling"

#: Built in family name for roof soffits
ROOF_SOFFIT_FAMILY_NAME = "Roof Soffit"

#: List of all Built in ceiling family names
BUILTIN_CEILING_TYPE_FAMILY_NAMES = [
    COMPOUND_CEILING_FAMILY_NAME,
    BASIC_CEILING_FAMILY_NAME,
    ROOF_SOFFIT_FAMILY_NAME,
]


def get_used_ceiling_type_ids(doc):
    """
    Gets all used ceiling type ids.
    Filters by category.
    Used: at least one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing used ceiling types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, rCeiling.get_all_ceiling_type_ids_in_model_by_category, 1
    )
    return ids


def family_no_types_in_use(fam_type_ids, un_used_type_ids):
    """
    Compares two lists of ids. True if any id is not in unUsedTypeIds.
    TODO: check for more generic list comparison and remove this function.

    :param fam_type_ids: List of family type ids to check.
    :type fam_type_ids: List of Autodesk.Revit.DB.ElementId
    :param un_used_type_ids: Reference list of ids.
    :type un_used_type_ids: List of Autodesk.Revit.DB.ElementId
    :return: True if any id from famTypeIds is not in unUsedTypeIds.
    :rtype: bool
    """

    match = True
    for fam_type_id in fam_type_ids:
        if fam_type_id not in un_used_type_ids:
            match = False
            break
    return match


# -------------------------------- In place ceiling types -------------------------------------------------------


def get_unused_non_in_place_ceiling_type_ids_to_purge(doc):
    """
    Gets all unused ceiling type id's.
    - Roof Soffit
    - Compound Ceiling
    - Basic Ceiling
    This method can be used to safely delete unused ceiling types:
    In the case that no ceiling instance using any of the types is placed this will return all but one type id since\
        Revit requires at least one ceiling type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing not used ceiling types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    # get unused type ids
    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, rCeiling.get_all_ceiling_type_ids_in_model_by_class, 0
    )
    # make sure there is at least on ceiling type per system family left in model
    ceiling_types = rCeilingTypeSort.sort_ceiling_types_by_family_name(doc)
    for key, value in ceiling_types.items():
        if key in BUILTIN_CEILING_TYPE_FAMILY_NAMES:
            if family_no_types_in_use(value, ids) == True:
                # remove one type of this system family from unused list
                ids.remove(value[0])
    return ids


def get_used_in_place_ceiling_type_ids(doc):
    """
    Gets all used in place ceiling type ids in the model.
    Used: at least one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing used in place ceiling types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, rCeiling.get_all_in_place_ceiling_type_ids_in_model, 1
    )
    return ids


def get_unused_in_place_ceiling_type_ids(doc):
    """
    Gets all unused in place ceiling type ids in the model.
    Unused: Not one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing unused in place ceiling types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, rCeiling.get_all_in_place_ceiling_type_ids_in_model, 0
    )
    return ids


def get_unused_in_place_ceiling_ids_for_purge(doc):
    """
    Gets symbol(type) ids and family ids (when no type is in use) of in place ceiling families which can be safely deleted from the model.
    This method can be used to safely delete unused in place ceiling types. There is no requirement by Revit to have at least one\
        in place ceiling definition in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing unused in place ceiling types and families.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = rFamPurge.get_unused_in_place_ids_for_purge(
        doc, get_unused_in_place_ceiling_type_ids
    )
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions to swap instances of a duplicated family to instances of the original family in a revit document.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This function will identify duplicated families in the current Revit document based on their names. It will then swap all instances of the duplicated families with instances of the original family, ensuring that the document maintains consistency and avoids redundancy.

Potential issues: 

- the source family does not include the type of the duplicate family

"""


#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Common import transaction as rTran
from duHast.Utilities.Objects import result as res
from family_find_duplicate import *
from duHast.UI.Objects.ProgressBase import ProgressBase
from duHast.Revit.Family.Data.Objects.family_directive_swap_instances_of_type import FamilyDirectiveSwap
from duHast.Revit.Family.family_swap_instances_of_types import _swap_loaded_family_instances
from duHast.Revit.Family.family_functions import get_name_and_category_to_family_dict

from Autodesk.Revit.DB import Element, ElementId, Transaction

DEBUG = False

def get_swap_directive(doc, source_type_id, target_type_id):
    # get the family types
    source_family_type = doc.GetElement(ElementId(source_type_id))
    target_family_type = doc.GetElement(ElementId(target_type_id))

    # get the source family
    source_family = source_family_type.Family
    source_family_category_name = source_family.FamilyCategory.Name
    
    directive = FamilyDirectiveSwap(
        name=Element.Name.GetValue(source_family),
        category=source_family_category_name,
        source_type_name=Element.Name.GetValue(source_family_type),
        target_family_name=Element.Name.GetValue(target_family_type.Family),
        target_family_type_name = Element.Name.GetValue(target_family_type)
    )

    if DEBUG:
        print ("Created directive: \n{}".format(directive.formatted_indented_str(1,"...")))
    
    return directive

def swap_family_instances_of_duplicates(doc, progress_callback=None):
    """
    Entry point for this module. 

    :return:
        Result class instance.

        - result.status. True if a single families was swapped successfully, otherwise False.
        - result.message will contain each swap message
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message in format: 'Failed to swap family from :' + currentName + ' to ' + newName
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    # check callback class
    if progress_callback and isinstance(progress_callback, ProgressBase) == False:
        raise TypeError(
            "progress_callback needs to be inherited from ProgressBase. Got : {} instead.".format(
                type(progress_callback)
            )
        )
    
    try:
        # attempt to find duplicate families
        duplicate_families = find_duplicate_families(doc)
        
        if DEBUG:
            print ("there...{}".format(len(duplicate_families)))
            for fam_id, families in duplicate_families.items():
                print ("fam_id: {}: {}".format(fam_id, families))

        # check if any duplicates found
        if len(duplicate_families) == 0:
            return_value.update_sep(True, "No duplicate families found.", [])
            return return_value
        
        
        # list to hold all swap directives
        swap_directives = []

        # find duplicate types in families
        for fam_id, families in duplicate_families.items():
            
            if DEBUG:
                print ("Processing family id: {}".format(fam_id))
            
            # get the target family
            target_family = doc.GetElement(fam_id)

            if DEBUG:
                print ("target_family: {}".format(target_family.Name))

            # get the duplicate families
            for duplicate_family in families:

                if DEBUG:
                    print ("duplicate_family: {}".format(duplicate_family.Name))
                
                # find matching types between families
                target_mapper = find_matching_types_between_families (duplicate_family, target_family)

                # if there are no matching types, skip to next duplicate family
                if len(target_mapper) == 0:
                    msg = "No matching types found between duplicate family: {} and target family: {}. Skipping.".format(duplicate_family.Name, target_family.Name)
                    if DEBUG:
                        print (msg)
                    return_value.append_message(msg)
                    continue
            
                # set up swap directives
                for dup_type_id, target_type_id in target_mapper.items():
                    
                    # build directive
                    directive = get_swap_directive(doc, dup_type_id, target_type_id)

                    # add to overall list of directives
                    swap_directives.append(directive)
                    if DEBUG:
                        print ("Added swap directive: \n{}".format(directive.formatted_indented_str(1,"...")))

        if DEBUG:
            print ("Total swap directives: {}".format(len(swap_directives)))
        
        # swap the instances
        if len(swap_directives)==0:
            return_value.update_sep(True, "No matching family types found between duplicate families and target families. No swaps performed.", [])
            return return_value
        
        # get all family in file
        families = get_name_and_category_to_family_dict(doc)

        # perform swap
        swap_result = _swap_loaded_family_instances(doc, swap_directives, families, progress_callback)

        print (swap_result)

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to swap family instances due to exception: {}".format(e)
        )
    
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to multi reference annotation. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from Autodesk.Revit.DB import (
    FilteredElementCollector,
    MultiReferenceAnnotationType,
    MultiReferenceAnnotation,
)
from duHast.Revit.Common.common import (
    get_ids_from_element_collector,
    get_similar_type_families_by_type,
)


def get_all_multi_ref_annotation_types(doc):
    """
    Gets all multi reference annotation types in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of MultiReferenceAnnotationType
    :rtype: Autodesk.Revit.DB.FilteredElementCollector of MultiReferenceAnnotationType
    """

    return FilteredElementCollector(doc).OfClass(MultiReferenceAnnotationType)


def get_all_multi_ref_annotation_type_ids(doc):
    """
    Gets all multi reference annotation type ids in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing multi reference Annotation Types
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col = FilteredElementCollector(doc).OfClass(MultiReferenceAnnotationType)
    ids = get_ids_from_element_collector(col)
    return ids


def get_all_multi_ref_annotation_elements(doc):
    """
    Gets all multi reference annotation elements in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of MultiReferenceAnnotation
    :rtype: Autodesk.Revit.DB.FilteredElementCollector of MultiReferenceAnnotation
    """

    return FilteredElementCollector(doc).OfClass(MultiReferenceAnnotation)


def get_all_similar_multi_reference_anno_types(doc):
    """
    Gets all multi reference annotation types using get similar types in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: list in format [[multi refType, [element ids of similar multi ref types, ...]]]
    :rtype: List [[Autodesk.Revit.DB.ElementType, Autodesk.Revit.DB.ElementId, Autodesk.Revit.DB.ElementId,...],]
    """

    multi_reference_anno_types = get_similar_type_families_by_type(
        doc, get_all_multi_ref_annotation_types
    )
    return multi_reference_anno_types



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Project parameters related methods
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


def get_project_parameter_bindings(doc):
    """
    Get all project parameter bindings in the document.
    
    :param doc: Current Revit document
    :type doc: Autodesk.Revit.DB.Document
    
    :return: List of parameter bindings
    :rtype: list of Autodesk.Revit.DB.ParameterBinding
    """
    
    map = doc.ParameterBindings
    it = map.ForwardIterator()
    it.Reset()

    result = []  

    while it.MoveNext():
        
        binding = it.Current  

        result.append(binding)

    return result


def get_project_parameter_definitions(doc):
    """
    Get all project parameter definitions in the document.
    
    :param doc: Current Revit document
    :type doc: Autodesk.Revit.DB.Document
    
    :return: List of parameter bindings
    :rtype: list of Autodesk.Revit.DB.ParameterBinding
    """
    
    map = doc.ParameterBindings
    it = map.ForwardIterator()
    it.Reset()

    result = []  

    while it.MoveNext():
        definition = it.Key
       
        result.append(definition)

    return result


def get_project_parameter_definition_by_name(doc,parameter_name):
    """
    Get a project parameter definition by name.
    
    :param doc: Current Revit document
    :type doc: Autodesk.Revit.DB.Document
    
    :param parameter_name: Name of the parameter to search for
    :type parameter_name: str
    
    :return: Parameter definition if found, None otherwise
    :rtype: Autodesk.Revit.DB.Definition or None
    """
    
    definitions = get_project_parameter_definitions(doc)
    
    for definition in definitions:
        if definition.Name == parameter_name:
            return definition
    
    return None


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit view filters storage reading from file. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Utilities.files_json import read_json_data_from_file
from duHast.Revit.Views.Objects.Data.view_filter import ViewFilter

from duHast.Utilities.Objects.result import Result

DEBUG = False

def read_filter_storage_from_file(file_path, node_name=""):

    """
    Read the view filter storage from file.

    :param file_path: Full path to the file including file name.
    :type file_path: str

    :return: Result message and list of view filter storage objects.
    :rtype: tuple (result message, list of view filter storage objects)
    """

    return_value = Result()

    try:
        # attempt to read data from file
        data_result = read_json_data_from_file(file_path)

        # check if read was successful
        if not data_result.status:
            return_value.update(data_result)
            return return_value

        # check if any data was found
        if len(data_result.result) == 0:
            return_value.update_sep(False, "No view filter storage data found in file.")
            return return_value
        
        data = data_result.result[0]
        
        # check if node name exists
        if node_name not in data:
            return_value.update_sep(False, "No view filter storage data found in file for node: {}.".format(node_name))
            print(data)
            return return_value

        # get a list of view filter storage objects
        view_filter_storage_data = data[node_name]

        if not isinstance(view_filter_storage_data, list):
            return_value.update_sep(False, "Filter storage data found in file for node: {} should be a list. Got {}.".format(node_name, type(view_filter_storage_data)))
            return return_value
        
        # loop over all view filter storage entries and create objects
        for view_filter_entry in view_filter_storage_data:
            if DEBUG:
                print("view filter entry in json: {}\n".format(view_filter_entry))
            # attempt to create view filter storage object
            try:
                view_filter = ViewFilter(j=view_filter_entry)
                if DEBUG:
                    print("Created view filter storage object: {}\n".format(view_filter))
                return_value.result.append(view_filter)
            except Exception as e:
                return_value.update_sep(False, "Failed to create view filter storage object from data in file for node: {}. {}".format(node_name, str(e)))
                

        return_value.update_sep(True, "Successfully read view filter storage from file.")
    
    except Exception as e:
        return_value.update_sep(False, "Failed to read view filter storage from file. {}".format(str(e)))
       

    return return_value




# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import clr
import os
import sys

from duHast.Utilities.Objects.result import Result

from System.Collections.Generic import List

def load_net_dll_path(dlls_to_load):
    """
    Loads dlls from the bin folder of the duHast extension.
    Useful if a dll is not loaded through the startup script of a pyRevit extension.

    expects the bin folder to be in the sys.path
    which means it needs to be located in the root folder of the extension:
    YourExtension.extension\\bin

    :param dlls_to_load: List of dlls to load.
    :type dlls_to_load: list
    :return: Result object with status and message.
    :rtype: Result
    """

    # set up a status tracker
    return_value = Result()

    try:

        # type checking
        if not isinstance(dlls_to_load, list):
            return_value.update_sep(
                False, "dlls_to_load must be a list, got {} instead.".format(type(dlls_to_load))
            )
            return return_value

        # get all the bin paths from the sys.path
        lib_paths = [path for path in sys.path if path.endswith("bin")]

        if not lib_paths or len(lib_paths) == 0:
            return_value.update_sep(
                False, "No bin paths found in sys.path."
            )

            for p in sys.path:
                return_value.append_message("sys.path entry: {}".format(p))

            return return_value
       
        return_value.append_message("Found lib paths: {}".format(lib_paths))
        
        # iterate over the dlls to load
        for dll in dlls_to_load:
            found_match = False
            # check if the dll exists in the lib paths
            for p in lib_paths:
                # build the actual path
                dll_path = os.path.join(p, dll)
                
                # check if path exists
                if os.path.exists(dll_path):
                    try:
                        # add the dll to the clr
                        clr.AddReferenceToFileAndPath(dll_path)
                        return_value.append_message("Loaded dll: {}".format(dll_path))
                        found_match = True
                        break
                    except Exception as e:
                        return_value.update_sep(
                            False, "Failed to load dll: {}. Error: {}".format(dll_path, e)
                        )
                        
            if not found_match:
                return_value.update_sep(
                    False, "DLL {} not found in any lib path.".format(dll)
                )
    except Exception as e:
        # handle any exceptions that occur during the export process
        message = "An error occurred while processing attempting to load dlls: {}".format(e)
        return_value.update_sep(
            False, message
        )


    return return_value